package geometry;

public class Geometry {
	
	public static final int MERGE = 0;
	public static final int SPLIT = 1;
	public static final int END = 2;
	public static final int START = 3;
	public static final int REGULAR = 4;
	
	public static final double ERROR = 1e-6;

	/** 
	 * Returns twice the area of the triangle abc. It is positive if c is left of ab, negative otherwise.
	 */
	public static int area(iPoint a, iPoint b, iPoint c){
		return (b.getX() - a.getX())*(c.getY() - a.getY()) - (c.getX() - a.getX())*(b.getY() - a.getY());
	}
	
	public static double area(Ponto a, Ponto b, Ponto c){
		return (b.getX() - a.getX())*(c.getY() - a.getY()) - (c.getX() - a.getX())*(b.getY() - a.getY());
	}
	
	/**
	 * Returns twice the area of the polygon.
	 */
	public static int area(Polygon p){
		int sum = 0;
		
		Vertex head = p.getHead();
		Vertex v = head.getVNext();
		
		do {
			sum += area(head.getPoint(), v.getPoint(), v.getVNext().getPoint());
			
			v = v.getVNext();
		} while(v.getVNext() != head);
		
		return sum;
	}
	
	/**
	 * Determines whether the point c is left of the segment ab. 
	 * @return true if c is left of ab, false otherwise.
	 */
	public static boolean left(iPoint a, iPoint b, iPoint c){	
		return (area(a, b, c) > 0);
	}
	
	/**
	 * Determines whether the point c is left of the segment ab. 
	 * @return true if c is left of ab, false otherwise.
	 */
	public static boolean left(Vertex a, Vertex b, Vertex c){	
		return (area(a.getPoint(), b.getPoint(), c.getPoint()) > 0);
	}
	
	public static boolean left(Ponto a, Ponto b, Ponto c){	
		return (area(a, b, c) > 0);
	}
	
	/**
	 * Determines whether the point c is left of or on the segment ab. 
	 * @return true if c is left of or on ab, false otherwise.
	 */
	public static boolean leftOn(iPoint a, iPoint b, iPoint c){	
		return (area(a, b, c) >= 0);
	}
	
	public static boolean leftOn(Ponto a, Ponto b, Ponto c){	
		return (area(a, b, c) >= 0);
	}
	
	/**
	 * Determines whether the point c is left of or on the segment ab. 
	 * @return true if c is left of or on ab, false otherwise.
	 */
	public static boolean leftOn(Vertex a, Vertex b, Vertex c){	
		return (area(a.getPoint(), b.getPoint(), c.getPoint()) >= 0);
	}
	
	/**
	 * Determines whether the points a, b and c are collinear. 
	 * @return true if the three points are collinear, false otherwise.
	 */
	public static boolean collinear(iPoint a, iPoint b, iPoint c){
		return (area(a, b, c) == 0);
	}
	
	public static boolean collinear(Ponto a, Ponto b, Ponto c){
		return ( Math.abs(area(a, b, c)) < ERROR);
	}

	/**
	 * Decides if the segments ab and cd intersects properly.
	 * @param a Point a of segment ab.
	 * @param b Point b of segment ab.
	 * @param c Point c of segment cd.
	 * @param d Point d of segment cd.
	 * @return true if ab intersects cd properly, false otherwise.
	 */
	public static boolean intersectProper(iPoint a, iPoint b, iPoint c, iPoint d){
		
		// Removes improper cases
		if(collinear(a, b, c) ||
		   collinear(a, b, d) ||
		   collinear(a, c, d) ||
		   collinear(b, c, d) )
			return false;
		
		return ((left(a, b, c)) != (left(a, b, d))) && ((left(c, d, a)) != (left(c, d, b)));
	}
	
	public static boolean intersectProper(Ponto a, Ponto b, Ponto c, Ponto d){

		// Removes improper cases
		if(collinear(a, b, c) ||
				collinear(a, b, d) ||
				collinear(a, c, d) ||
				collinear(b, c, d) )
			return false;

		return ((left(a, b, c)) != (left(a, b, d))) && ((left(c, d, a)) != (left(c, d, b)));
	}
	
	public static boolean intersectProper(LineSegment l1, LineSegment l2){
		return intersectProper(l1.getP1(), l1.getP2(), l2.getP1(), l2.getP2());		
	}

	/**
	 * Decides if point c is 'between' the segment ab
	 * @param a Point a of segment ab
	 * @param b Point b of segment ab
	 * @param c Point to check
	 * @return true if c is between the segment ab, false otherwise
	 */
	public static boolean between(iPoint a, iPoint b, iPoint c){
		
		if(!collinear(a, b, c)){
	        return false;
		}

	    // If ab is not vertical, checks the x coordinate, otherwise checks the y coordinate
	    if(a.getX() == b.getX())
	        return ( (a.getX() <= c.getX()) && (c.getX() <= b.getX()) ) || ( (a.getX() >= c.getX()) && (c.getX() >= b.getX()) );
	    else
	        return ( (a.getY() <= c.getY()) && (c.getY() <= b.getY()) ) || ( (a.getY() >= c.getY()) && (c.getY() >= b.getY()) );
		
	}
	
	/**
	 * Decides if the segments ab and cd intersects.
	 * @param a Point a of segment ab.
	 * @param b Point b of segment ab.
	 * @param c Point c of segment cd.
	 * @param d Point d of segment cd.
	 * @return true if ab intersects cd, false otherwise.
	 */
	public static boolean intersect(iPoint a, iPoint b, iPoint c, iPoint d)
	{
	    if(intersectProper(a, b, c, d))
	        return true;
	    else if( between(a, b, c) ||
	             between(a, b, d) ||
	             between(c, d, a) ||
	             between(c, d, b) )
	        return true;
	    return false;
	}
	
	/**
	 * Decides if the segments ab and cd intersects.
	 * @param a Vertex a of segment ab.
	 * @param b Vertex b of segment ab.
	 * @param c Vertex c of segment cd.
	 * @param d Vertex d of segment cd.
	 * @return true if ab intersects cd, false otherwise.
	 */
	public static boolean intersect(Vertex a, Vertex b, Vertex c, Vertex d)
	{
		return (intersect(a.getPoint(), b.getPoint(), c.getPoint(), d.getPoint()));
	}

	/**
	 * Decides if the segment is a non-intersecting diagonal of the polygon, internal or external.
	 * @param a Vertex a of the segment to be checked
	 * @param b Vertex b of the segment to be checked
	 * @return true if ab is a non-intersecting diagonal of the polygon p, internal or external, false otherwise.
	 */
	public static boolean diagonalInternalExternal(Polygon p, Vertex a, Vertex b)
	{
	    Vertex c, c1;

	    c = p.getHead();
	    
	    // For each edge (c, c1)
	    do {
	        c1 = c.getVNext();
	        
	        if( (c != a) && (c1 != a) && (c != b) && (c1 != b) && intersect(a, b, c, c1)){
	            return false;
	        }
	        
	        c = c.getVNext();
	        
	    } while(c != p.getHead());
	    return true;
	}

	/**
	 * Decides if the segment is an internal or external diagonal.
	 * @param a Vertex a of segment ab.
	 * @param b Vertex b of segment ab.
	 * @return Returns true if ab is an internal diagonal, false otherwise.
	 */
	public static boolean inCone(Vertex a, Vertex b)
	{
		
		// aPrev, a, aNext consecutive vertices
	    Vertex aPrev, aNext; 

	    aNext = a.getVNext();
	    aPrev = a.getVPrev();

	    // If it is convex
	    if( leftOn(a, aNext, aPrev)) {
	        return left(a, b, aPrev) && left(b, a, aNext);
	    } else { // Reflex
	        return !(leftOn(a, b, aNext) && leftOn(b, a, aPrev));
	    }
	}

	/**
	 * Decides if segment is a valid diagonal of the polygon.
	 * @param p Polygon p.
	 * @param a End-point a of segment ab.
	 * @param b End-point b of segment ab.
	 * @return true if ab is a valid diagonal of the polygon p.
	 */
	public static boolean diagonal(Polygon p, Vertex a, Vertex b)
	{
	    return inCone(a, b) && inCone(b, a) && diagonalInternalExternal(p, a, b);
	}
	
	/**
	 * Decides if the vertex is an ear.
	 */
	public static boolean isEar(Polygon p, Vertex v){
		return diagonal(p, v.getVPrev(), v.getVNext());
	}
	
	/**
	 * Initialize every vertex ear status. 
	 * @param p Polygon.
	 */
	public static void initializeEarStatus(Polygon p){
		
		Vertex v = p.getHead();
		
		do {
			v.setEar(isEar(p, v));
			
			v = v.getVNext();
		} while(v != p.getHead());
	}

	/**
	 * Removes an ear from the polygon.
	 * @param poly Polygon.
	 * @param v Vertex to be removed. Must be an ear.
	 */
	public static void removeEar(Polygon poly, Vertex v) {
		
		if(!v.isEar()){
			return;
		}
		
		poly.setHead(v.getVNext());
		v.getVNext().setVPrev(v.getVPrev());
		v.getVPrev().setVNext(v.getVNext());		
	}

	public static void initializeVertexStatus(Polygon poly) {
		
		Vertex v = poly.getHead();
		
		do {
			v.setStatus(status(poly, v));
			
			v = v.getVNext();
		} while(v != poly.getHead());
		
	}

	private static int status(Polygon poly, Vertex v) {
		if(isMerge(v)){
			return MERGE;
		} else if(isSplit(v)){
			return SPLIT;
		} else if(isRegular(v)){
			return REGULAR;
		} else if(isStart(v)){
			return START;
		} else if(isEnd(v)){
			return END;
		}	
		return -1;
	}

	private static boolean isEnd(Vertex v) {
		return (v.getVNext().getPoint().getY() > v.getPoint().getY()
				&& v.getVPrev().getPoint().getY() > v.getPoint().getY()
				&& left(v.getVPrev().getPoint(), v.getPoint(), v.getVNext().getPoint()));
	}

	private static boolean isStart(Vertex v) {
		return (v.getVNext().getPoint().getY() < v.getPoint().getY()
				&& v.getVPrev().getPoint().getY() < v.getPoint().getY()
				&& left(v.getVPrev().getPoint(), v.getPoint(), v.getVNext().getPoint()));
	}

	private static boolean isRegular(Vertex v) {
		return ( (v.getVNext().getPoint().getY() < v.getPoint().getY())
				!= (v.getVPrev().getPoint().getY() < v.getPoint().getY()) );
	}

	private static boolean isSplit(Vertex v) {
		return (v.getVNext().getPoint().getY() < v.getPoint().getY()
				&& v.getVPrev().getPoint().getY() < v.getPoint().getY()
				&& !left(v.getVPrev().getPoint(), v.getPoint(), v.getVNext().getPoint()));
	}

	private static boolean isMerge(Vertex v) {
		return (v.getVNext().getPoint().getY() > v.getPoint().getY()
				&& v.getVPrev().getPoint().getY() > v.getPoint().getY()
				&& !left(v.getVPrev().getPoint(), v.getPoint(), v.getVNext().getPoint()));
	}
	
	/**
	 * Segments must be disjoint (no intersection)
	 * Returns if segment b is left (X-axis meaning) of a
	 */
	public static boolean segmentIsLeftOf(Edge a, Edge b){
		if(left(a.getEndPoint1(), a.getEndPoint2(), b.getEndPoint1()) == left(a.getEndPoint1(), a.getEndPoint2(), b.getEndPoint2())){
			return (left(a.getEndPoint1(), a.getEndPoint2(), b.getEndPoint1()));
		} else {
			return (!left(b.getEndPoint1(), b.getEndPoint2(), a.getEndPoint1()));
		}
	}
	
	public static void initializeChain(Polygon p){
		
		Vertex highest = p.getHighest();
		Vertex lowest = p.getLowest();
		
		for(Vertex v = highest; v != lowest ; v = v.getVNext()){
			v.setChain(1);
		}
		
		for(Vertex v = highest; v != lowest ; v = v.getVPrev()){
			v.setChain(0);
		}
		
		lowest.setChain(0);	
		
	}
	
}
